from rest_framework.permissions import SAFE_METHODS

from arkindex.ponos.models import Task
from arkindex.project.mixins import ProcessACLMixin
from arkindex.project.permissions import IsAuthenticated, IsVerified
from arkindex.users.models import Role


def require_agent_or_admin(request, view):
    return getattr(request.user, "is_admin", False) or getattr(request.user, "is_agent", False)


def require_task(request, view):
    return getattr(request.user, "is_admin", False) or isinstance(request.auth, Task)


def require_agent_or_task(request, view):
    return (
        getattr(request.user, "is_agent", False)
        or require_task(request, view)
    )


def require_verified_or_agent(request, view):
    return getattr(request.user, "verified_email", False) or require_agent_or_admin(request, view)


class IsAgent(IsAuthenticated):
    """
    Only allow Ponos agents and admins.
    """
    checks = IsAuthenticated.checks + (require_agent_or_admin, )


class IsAgentOrTask(IsAuthenticated):
    """
    Only allow Ponos agents, tasks, and admins.
    """
    checks = IsAuthenticated.checks + (require_agent_or_task, )


class IsAgentOrReadOnly(IsAgent):
    """
    Restricts write access to Ponos agents and admins,
    leave read access to everyone without any authentication.
    """
    allow_safe_methods = True


class IsAssignedAgentOrReadOnly(IsAgentOrReadOnly):
    """
    Restricts write access to Ponos agents and admins,
    and restricts write access on tasks to admins and agents assigned to them.
    Leave read access to everyone without any authentication.
    """

    def has_object_permission(self, request, view, obj) -> bool:
        if (
            isinstance(obj, Task)
            and not getattr(request.user, "is_admin", False)
            and not request.user == obj.agent
        ):
            return request.method in SAFE_METHODS
        return super().has_object_permission(request, view, obj)


class IsAssignedAgentOrTaskOrReadOnly(ProcessACLMixin, IsAgentOrTask):
    """
    Restricts write access to Ponos agents, Ponos tasks, and admins, and allows read access to anyone.
    When checking object write permissions for a Ponos task, requires either a Ponos agent assigned to the task,
    or authentication as the task itself.
    """
    allow_safe_methods = True

    def has_object_permission(self, request, view, obj) -> bool:
        assert isinstance(obj, Task)

        if request.method in SAFE_METHODS:
            return True

        if isinstance(request.auth, Task):
            return obj == request.auth

        return (super().has_object_permission(request, view, obj) and obj.agent_id == request.user.id)


class IsTaskAdmin(ProcessACLMixin, IsVerified):
    """
    Allow instance admins and users with a verified email and admin access to the task's process.
    """

    def has_object_permission(self, request, view, task):
        # Add request to attributes for the ACL mixin to work with self.user
        self.request = request
        level = self.process_access_level(task.process)
        # process_access_level can return None if there is no access at all
        return level and level >= Role.Admin.value


class IsAgentOrTaskGuest(ProcessACLMixin, IsAuthenticated):
    """
    Allow admins, Ponos agents, users with a verified email and guest access to the task's process, the task itself,
    or another task belonging to a process the creator of which has access.
    """

    def has_object_permission(self, request, view, task):
        assert isinstance(task, Task)

        # Add request to attributes for the ACL mixin to work with self.user
        self.request = request

        return (
            task == request.auth
            or require_agent_or_admin(request, view)
            or (
                # Users with a verified email
                getattr(request.user, "verified_email", False)
                # process_access_level can return None if there is no access at all
                and (self.process_access_level(task.process) or 0) >= Role.Guest.value
            )
        )


class IsAgentOrArtifactGuest(IsAgentOrTaskGuest):
    """
    Permission to access an artifact, based on access to the artifact's task through IsAgentOrTaskGuest.
    """

    def has_object_permission(self, request, view, artifact):
        return super().has_object_permission(request, view, artifact.task)


class IsTask(IsAuthenticated):
    checks = (require_task, )


class IsVerifiedOrAgent(IsVerified):
    """
    Allows Ponos agents, instance admins and users with a verified email.
    """
    checks = (require_verified_or_agent, )
